@@ -1,36 +1,37 @@
-from dataclasses import dataclass
 from datetime import datetime
 from decimal import Decimal
-from typing import Union, List, Dict
 from functools import lru_cache
+from typing import Union, List, Dict, Any, Optional
 
 import pandas as pd
 from bs4 import Tag
+from pydantic import BaseModel
 from rich import box
 from rich.console import Group, Text
+from rich.panel import Panel
 from rich.table import Table
 
-from edgar._rich import repr_rich, df_to_rich_table
-from edgar.core import moneyfmt
-from edgar._xml import find_element, child_text, optional_decimal
+from edgar.core import moneyfmt, get_bool
+from edgar.funds import get_fund_information, FundSeriesAndContracts, Fund, get_fund
+from edgar.reference import cusip_ticker_mapping
+from edgar.richtools import repr_rich, df_to_rich_table
+from edgar.xmltools import find_element, child_text, optional_decimal
 
 __all__ = [
     "FundReport",
     "CurrentMetric",
-    "FUND_FORMS"
+    "NPORT_FORMS"
 ]
 
-FUND_FORMS = ["NPORT-P", "NPORT-EX"]
+NPORT_FORMS: List[str] = ["NPORT-P", "NPORT-EX"]
 
 
-@dataclass(frozen=True)
-class IssuerCredentials:
+class IssuerCredentials(BaseModel):
     cik: str
     ccc: str  # cik confirmation code
 
 
-@dataclass(frozen=True)
-class SeriesClassInfo:
+class SeriesClassInfo(BaseModel):
     series_id: str
     class_id: str
 
@@ -42,43 +43,46 @@ class SeriesClassInfo:
                        class_id=child_text(tag, "classId"))
 
 
-@dataclass(frozen=True)
-class FilerInfo:
+class FilerInfo(BaseModel):
     issuer_credentials: IssuerCredentials
-    series_class_info: SeriesClassInfo
+    series_class_info: Optional[SeriesClassInfo]
+
+    @property
+    def series_id(self):
+        return self.series_class_info.series_id if self.series_class_info else ""
+
+    @property
+    def class_id(self):
+        return self.series_class_info.class_id if self.series_class_info else ""
 
 
-@dataclass(frozen=True)
-class Header:
+class Header(BaseModel):
     submission_type: str
     is_confidential: bool
     filer_info: FilerInfo
 
 
-@dataclass(frozen=True)
-class GeneralInfo:
+class GeneralInfo(BaseModel):
     name: str
     cik: str
     file_number: str
-    lei: str
+    reg_lei: Optional[str]
     street1: str
-    street2: str
-    city: str
-    state: str
-    country: str
-    zip_or_postal_code: str
-    phone: str
-    series_name: str
-    series_lei: str
-    series_id: str
-    class_ids: List[str]
-    reg_period_end: datetime
-    reg_period_date: datetime
-    is_final_filing: bool
+    street2: Optional[str]
+    city: Optional[str]
+    state: Optional[str]
+    country: Optional[str]
+    zip_or_postal_code: Optional[str]
+    phone: Optional[str]
+    series_name: Optional[str]
+    series_lei: Optional[str]
+    series_id: Optional[str]
+    fiscal_year_end: Optional[str]
+    rep_period_date: Optional[str]
+    is_final_filing: Optional[bool]
 
 
-@dataclass(frozen=True)
-class PeriodType:
+class PeriodType(BaseModel):
     # These elements are part of B.3.c. Credit Spread Risk (SDV01, CR01 or CS01).
     # e.g. <intrstRtRiskdv01 period3Mon="0" period1Yr="0" period5Yr="0" period10Yr="0" period30Yr="0"/>
     period3Mon: Decimal
@@ -99,8 +103,7 @@ class PeriodType:
                        )
 
 
-@dataclass(frozen=True)
-class CurrentMetric:
+class CurrentMetric(BaseModel):
     currency: str
     intrstRtRiskdv01: PeriodType
     intrstRtRiskdv100: PeriodType
@@ -116,12 +119,17 @@ def datetime_or_na(value: str):
     return value if value == "N/A" else datetime.strptime(value, "%Y-%m-%d")
 
 
-@dataclass(frozen=True)
-class MonthlyTotalReturn:
-    class_id: str
-    return1: Decimal
-    return2: Decimal
-    return3: Decimal
+def format_date(date: Union[str, datetime]) -> str:
+    if isinstance(date, str):
+        return date
+    return date.strftime("%Y-%m-%d")
+
+
+class MonthlyTotalReturn(BaseModel):
+    class_id: Optional[str]
+    return1: Optional[Union[Decimal, str]]
+    return2: Optional[Union[Decimal, str]]
+    return3: Optional[Union[Decimal, str]]
 
     @classmethod
     def from_xml(cls, tag: Tag):
@@ -133,10 +141,9 @@ class MonthlyTotalReturn:
         )
 
 
-@dataclass(frozen=True)
-class RealizedChange:
-    net_realized_gain: Decimal
-    net_unrealized_appreciation: Decimal
+class RealizedChange(BaseModel):
+    net_realized_gain: Optional[Union[Decimal, str]]
+    net_unrealized_appreciation: Optional[Union[Decimal, str]]
 
     @classmethod
     def from_xml(cls,
@@ -148,11 +155,10 @@ class RealizedChange:
             )
 
 
-@dataclass(frozen=True)
-class MonthlyFlow:
-    redemption: Decimal
-    reinvestment: Decimal
-    sales: Decimal
+class MonthlyFlow(BaseModel):
+    redemption: Optional[Union[Decimal, str]]
+    reinvestment: Optional[Union[Decimal, str]]
+    sales: Optional[Union[Decimal, str]]
 
     @classmethod
     def from_xml(cls,
@@ -165,48 +171,45 @@ class MonthlyFlow:
             )
 
 
-@dataclass(frozen=True)
-class ReturnInfo:
+class ReturnInfo(BaseModel):
     monthly_total_returns: List[MonthlyTotalReturn]
     other_mon1: RealizedChange
     other_mon2: RealizedChange
     other_mon3: RealizedChange
 
 
-@dataclass(frozen=True)
-class FundInfo:
+class FundInfo(BaseModel):
     total_assets: Decimal
     total_liabilities: Decimal
-    net_assets: Decimal
-    assets_attr_misc_sec: Decimal
-    assets_invested: Decimal
-    amt_pay_one_yr_banks_borr: Decimal
-    amt_pay_one_yr_ctrld_comp: Decimal
-    amt_pay_one_yr_oth_affil: Decimal
-    amt_pay_one_yr_other: Decimal
-    amt_pay_aft_one_yr_banks_borr: Decimal
-    amt_pay_aft_one_yr_ctrld_comp: Decimal
-    amt_pay_aft_one_yr_oth_affil: Decimal
-    amt_pay_aft_one_yr_other: Decimal
-    delay_deliv: Decimal
-    stand_by_commit: Decimal
-    liquidity_pref: Decimal
-    cash_not_report_in_cor_d: Decimal
+    net_assets: Optional[Decimal]
+    assets_attr_misc_sec: Optional[Decimal]
+    assets_invested: Optional[Decimal]
+    amt_pay_one_yr_banks_borr: Optional[Decimal]
+    amt_pay_one_yr_ctrld_comp: Optional[Decimal]
+    amt_pay_one_yr_oth_affil: Optional[Decimal]
+    amt_pay_one_yr_other: Optional[Decimal]
+    amt_pay_aft_one_yr_banks_borr: Optional[Decimal]
+    amt_pay_aft_one_yr_ctrld_comp: Optional[Decimal]
+    amt_pay_aft_one_yr_oth_affil: Optional[Decimal]
+    amt_pay_aft_one_yr_other: Optional[Decimal]
+    delay_deliv: Optional[Decimal]
+    stand_by_commit: Optional[Decimal]
+    liquidity_pref: Optional[Decimal]
+    cash_not_report_in_cor_d: Optional[Decimal]
     current_metrics: Dict[str, CurrentMetric]
-    credit_spread_risk_investment_grade: PeriodType
-    credit_spread_risk_non_investment_grade: PeriodType
-    is_non_cash_collateral: bool
-    return_info: ReturnInfo
-    monthly_flow1: MonthlyFlow
-    monthly_flow2: MonthlyFlow
-    monthly_flow3: MonthlyFlow
+    credit_spread_risk_investment_grade: Optional[PeriodType]
+    credit_spread_risk_non_investment_grade: Optional[PeriodType]
+    is_non_cash_collateral: Optional[bool]
+    return_info: Optional[ReturnInfo]
+    monthly_flow1: Optional[MonthlyFlow]
+    monthly_flow2: Optional[MonthlyFlow]
+    monthly_flow3: Optional[MonthlyFlow]
 
 
-@dataclass(frozen=True)
-class DebtSecurity:
-    maturity_date: datetime
+class DebtSecurity(BaseModel):
+    maturity_date: Union[datetime, str]
     coupon_kind: str
-    annualized_rate: Decimal
+    annualized_rate: Optional[Decimal]
     is_default: bool
     are_instrument_payents_in_arrears: bool
     is_paid_kind: bool
@@ -229,11 +232,10 @@ class DebtSecurity:
             )
 
 
-@dataclass(frozen=True)
-class SecurityLending:
-    is_cash_collateral: bool
-    is_non_cash_collateral: bool
-    is_loan_by_fund: bool
+class SecurityLending(BaseModel):
+    is_cash_collateral: Optional[str]
+    is_non_cash_collateral: Optional[str]
+    is_loan_by_fund: Optional[str]
 
     @classmethod
     def from_xml(cls, tag):
@@ -245,11 +247,10 @@ class SecurityLending:
             )
 
 
-@dataclass(frozen=True)
-class Identifiers:
-    ticker: str
-    isin: str
-    other: str
+class Identifiers(BaseModel):
+    ticker: Optional[str]
+    isin: Optional[str]
+    other: Dict
 
     @classmethod
     def from_xml(cls,
@@ -267,27 +268,26 @@ class Identifiers:
             return cls(ticker=ticker, isin=isin, other=other)
 
 
-@dataclass(frozen=True)
-class InvestmentOrSecurity:
+class InvestmentOrSecurity(BaseModel):
     name: str
     lei: str
     title: str
     cusip: str
     identifiers: Identifiers
-    balance: Decimal
-    units: str
-    desc_other_units: str
-    currency_code: str
+    balance: Optional[Decimal]
+    units: Optional[str]
+    desc_other_units: Optional[str]
+    currency_code: Optional[str]
     value_usd: Decimal
     pct_value: Decimal
-    payoff_profile: str
-    asset_category: str
-    issuer_category: str
-    investment_country: str
+    payoff_profile: Optional[str]
+    asset_category: Optional[str]
+    issuer_category: Optional[str]
+    investment_country: Optional[str]
     is_restricted_security: bool
-    fair_value_level: str
-    debt_security: DebtSecurity
-    security_lending: SecurityLending
+    fair_value_level: Optional[str]
+    debt_security: Optional[DebtSecurity]
+    security_lending: Optional[SecurityLending]
 
     @property
     def ticker(self):
@@ -299,25 +299,35 @@ class InvestmentOrSecurity:
 
 
 class FundReport:
+    """
+    Form N-PORT-P is a form filed with the SEC by mutual funds to report their monthly portfolio holdings to the SEC.
+    """
 
     def __init__(self,
                  header: Header,
                  general_info: GeneralInfo,
                  fund_info: FundInfo,
-                 investments: List[InvestmentOrSecurity]):
+                 investments: List[InvestmentOrSecurity],
+                 series_and_contracts: FundSeriesAndContracts = None):
         self.header = header
         self.general_info: GeneralInfo = general_info
         self.fund_info: FundInfo = fund_info
-        self.investments = investments
+        self.investments: List[InvestmentOrSecurity] = investments
+        self.series_and_contracts: FundSeriesAndContracts = series_and_contracts
 
     def __str__(self):
-        return (f"{self.name} {self.general_info.reg_period_date} - {self.general_info.reg_period_end}"
+        return (f"{self.name} {self.general_info.rep_period_date} - {self.general_info.fiscal_year_end}"
                 )
 
     @property
     def name(self):
         return f"{self.general_info.name} - {self.general_info.series_name}"
 
+    @property
+    @lru_cache(maxsize=2)
+    def fund(self) -> Fund:
+        return get_fund(self.general_info.series_id)
+
     @property
     def has_investments(self):
         return len(self.investments) > 0
@@ -329,7 +339,11 @@ class FundReport:
         """
         if len(self.investments) == 0:
             return pd.DataFrame(columns=['name', 'title', 'cusip', 'ticker', 'balance', 'units'])
-        return pd.DataFrame(
+
+        # This is for adding Ticker to the investments in case it is None
+        cusip_mapping = cusip_ticker_mapping(allow_duplicate_cusips=False)
+
+        investment_df = pd.DataFrame(
             [{
                 "name": investment.name,
                 "title": investment.title,
@@ -360,9 +374,25 @@ class FundReport:
             ]
         ).sort_values(['value_usd', 'name', 'title'], ascending=[False, True, True]).reset_index(drop=True)
 
+        # Step 1: Map CUSIP to Ticker using the cusip_mapping
+        mapped_tickers = investment_df.cusip.map(cusip_mapping.Ticker)
+
+        # Step 2: Fill NaN values in the ticker column with mapped tickers
+        investment_df['ticker'] = investment_df['ticker'].astype(str).fillna(mapped_tickers).fillna("")
+
+        return investment_df
+
     @classmethod
-    def from_xml(cls,
-                 xml: Union[str, Tag]):
+    def from_filing(cls, filing):
+        xml = filing.xml()
+        fund_report_dict = FundReport.parse_fund_xml(xml)
+        # Parse ticker, fund, series information from the filing header
+        fund_report_dict['series_and_contracts'] = get_fund_information(filing.header)
+        return cls(**fund_report_dict)
+
+    @classmethod
+    def parse_fund_xml(cls,
+                       xml: Union[str, Tag]) -> Dict[str, Any]:
         root = find_element(xml, "edgarSubmission")
 
         # Get the header
@@ -402,7 +432,7 @@ class FundReport:
             name=child_text(general_info_tag, "regName"),
             cik=child_text(general_info_tag, "regCik"),
             file_number=child_text(general_info_tag, "regFileNumber"),
-            lei=child_text(general_info_tag, "regLei"),
+            reg_lei=child_text(general_info_tag, "regLei"),
             street1=child_text(general_info_tag, "regStreet1"),
             street2=child_text(general_info_tag, "regStreet2"),
             city=child_text(general_info_tag, "regCity"),
@@ -412,12 +442,10 @@ class FundReport:
             country=country,
             series_name=child_text(general_info_tag, "seriesName"),
             series_id=child_text(general_info_tag, "seriesId"),
-            class_ids=[child_text(tag) for tag in general_info_tag.find_all("class_id")],
             series_lei=child_text(general_info_tag, "seriesLei"),
-            reg_period_end=child_text(general_info_tag, "repPdEnd"),
-            reg_period_date=child_text(general_info_tag, "repPdDate"),
-            is_final_filing=child_text(general_info_tag, "isFinalFiling")
-
+            fiscal_year_end=child_text(general_info_tag, "repPdEnd"),
+            rep_period_date=child_text(general_info_tag, "repPdDate"),
+            is_final_filing=get_bool(child_text(general_info_tag, "isFinalFiling"))
         )
 
         # Fund info
@@ -482,55 +510,60 @@ class FundReport:
         investment_or_secs_tag = form_data_tag.find("invstOrSecs")
         if investment_or_secs_tag:
             investments_or_securities = []
-            for investment_or_sec_tag in investment_or_secs_tag.find_all("invstOrSec"):
+            for investment_tag in investment_or_secs_tag.find_all("invstOrSec"):
                 # issuer conditional
-                asset_conditional_tag = investment_or_sec_tag.find("assetConditional")
+                asset_conditional_tag = investment_tag.find("assetConditional")
                 if asset_conditional_tag:
                     asset_category = asset_conditional_tag.attrs.get("assetCat")
                 else:
-                    asset_category = child_text(investment_or_sec_tag, "assetCat")
+                    asset_category = child_text(investment_tag, "assetCat")
 
                 # issuer conditional
-                issuer_conditional_tag = investment_or_sec_tag.find("issuerConditional")
+                issuer_conditional_tag = investment_tag.find("issuerConditional")
                 if issuer_conditional_tag:
                     issuer_category = issuer_conditional_tag.attrs.get("issuerCat")
                 else:
-                    issuer_category = child_text(investment_or_sec_tag, "issuerCat")
-
-                    [
-
-                    ]
+                    issuer_category = child_text(investment_tag, "issuerCat")
 
                 investments_or_security = InvestmentOrSecurity(
-                    name=child_text(investment_or_sec_tag, "name"),
-                    lei=child_text(investment_or_sec_tag, "lei"),
-                    title=child_text(investment_or_sec_tag, "title"),
-                    cusip=child_text(investment_or_sec_tag, "cusip"),
-                    identifiers=Identifiers.from_xml(investment_or_secs_tag.find("identifiers")),
-                    balance=optional_decimal(investment_or_sec_tag, "balance"),
-                    units=child_text(investment_or_sec_tag, "units"),
-                    desc_other_units=child_text(investment_or_sec_tag, "descOthUnits"),
-                    currency_code=child_text(investment_or_sec_tag, "curCd"),
-                    value_usd=optional_decimal(investment_or_sec_tag, "valUSD"),
-                    pct_value=optional_decimal(investment_or_sec_tag, "pctVal"),
-                    payoff_profile=child_text(investment_or_sec_tag, "payoffProfile"),
+                    name=child_text(investment_tag, "name"),
+                    lei=child_text(investment_tag, "lei"),
+                    title=child_text(investment_tag, "title"),
+                    cusip=child_text(investment_tag, "cusip"),
+                    identifiers=Identifiers.from_xml(investment_tag.find("identifiers")),
+                    balance=optional_decimal(investment_tag, "balance"),
+                    units=child_text(investment_tag, "units"),
+                    desc_other_units=child_text(investment_tag, "descOthUnits"),
+                    currency_code=child_text(investment_tag, "curCd"),
+                    value_usd=optional_decimal(investment_tag, "valUSD"),
+                    pct_value=optional_decimal(investment_tag, "pctVal"),
+                    payoff_profile=child_text(investment_tag, "payoffProfile"),
                     asset_category=asset_category,
                     issuer_category=issuer_category,
-                    investment_country=child_text(investment_or_sec_tag, "invCountry"),
-                    is_restricted_security=child_text(investment_or_sec_tag, "isRestrictedSec") == "Y",
-                    fair_value_level=child_text(investment_or_sec_tag, "fairValLevel"),
-                    debt_security=DebtSecurity.from_xml(investment_or_sec_tag.find("debtSec")),
-                    security_lending=SecurityLending.from_xml(investment_or_sec_tag.find("securityLending"))
+                    investment_country=child_text(investment_tag, "invCountry"),
+                    is_restricted_security=child_text(investment_tag, "isRestrictedSec") == "Y",
+                    fair_value_level=child_text(investment_tag, "fairValLevel"),
+                    debt_security=DebtSecurity.from_xml(investment_tag.find("debtSec")),
+                    security_lending=SecurityLending.from_xml(investment_tag.find("securityLending"))
                 )
 
                 investments_or_securities.append(investments_or_security)
 
-        fund_report = FundReport(header=header,
-                                 general_info=general_info,
-                                 fund_info=fund_info,
-                                 investments=investments_or_securities)
+        # Get the fund Information from the filing header
 
-        return fund_report
+        return {'header': header,
+                'general_info': general_info,
+                'fund_info': fund_info,
+                'investments': investments_or_securities}
+
+    @property
+    def fund_info_table(self) -> Table:
+        fund_info_table = Table("Fund", "Series", "As Of Date", "Fiscal Year", box=box.SIMPLE)
+        fund_info_table.add_row(self.general_info.name,
+                                f"{self.general_info.series_name} {self.general_info.series_id or ''}",
+                                self.general_info.rep_period_date,
+                                self.general_info.fiscal_year_end)
+        return fund_info_table
 
     @property
     def fund_summary_table(self) -> Table:
@@ -539,13 +572,11 @@ class FundReport:
                                  "Liabilities",
                                  "Net Assets",
                                  "Investments",
-                                 "Period",
-                                 title="Fund Summary", title_style="bold deep_sky_blue1", box=box.SIMPLE)
+                                 title="Financials", title_style="bold deep_sky_blue1", box=box.SIMPLE)
         financials_table.add_row(moneyfmt(self.fund_info.total_assets, curr="$", places=0),
                                  moneyfmt(self.fund_info.total_liabilities, curr="$", places=0),
                                  moneyfmt(self.fund_info.net_assets, curr="$", places=0),
-                                 f"{len(self.investments)}",
-                                 f"{self.general_info.reg_period_date} - {self.general_info.reg_period_end}"
+                                 f"{len(self.investments)}"
                                  )
         return financials_table
 
@@ -606,19 +637,21 @@ class FundReport:
                            .assign(Name=lambda df: df.name,
                                    Title=lambda df: df.title,
                                    Cusip=lambda df: df.cusip,
+                                   Ticker=lambda df: df.ticker,
                                    Value=lambda df: df.value_usd.apply(moneyfmt, curr='$', places=0),
-                                   Percent=lambda df: df.pct_value.apply(moneyfmt, curr='', places=1),
+                                   Pct=lambda df: df.pct_value.apply(moneyfmt, curr='', places=1),
                                    Category=lambda df: df.issuer_category + " " + df.asset_category)
-                           ).filter(['Name', 'Title', 'Cusip', 'Category', 'Value', 'Percent'])
+                           ).filter(['Name', 'Title', 'Cusip', 'Ticker', 'Category', 'Value', 'Pct'])
         return df_to_rich_table(investments, title="Investments", title_style="bold deep_sky_blue1", max_rows=2000)
 
     def __rich__(self):
-        return Group(
-            Text(self.name, style="bold dark_sea_green4"),
+        title = f"{self.general_info.name} - {self.general_info.series_name} {self.general_info.rep_period_date}"
+        return Panel(Group(
             self.fund_summary_table,
             self.metrics_table,
             self.credit_spread_table,
             self.investments_table
+        ), title=title, subtitle=title
         )
 
     def __repr__(self):
